import sys
import warnings

import pytest

import solcx
from solcx.exceptions import SolcInstallationError, UnexpectedVersionError, UnexpectedVersionWarning


def test_validate_installation(mocker, install_mock, solc_binary, install_path):
    version = solcx.wrapper.get_solc_version(solc_binary)
    version_str_patch = mocker.patch("solcx.wrapper.get_version_str_from_solc_binary")
    version_str_patch.return_value = f"{version}"

    # This should not warn!
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        solcx.install_solc()

    assert install_path.exists()


def test_validate_installation_wrong_version(mocker, install_mock, install_path):
    patch = mocker.patch("solcx.wrapper.get_version_str_from_solc_binary")
    patch.return_value = "0.0.0"

    with pytest.raises(UnexpectedVersionError):
        solcx.install_solc()

    assert not install_path.exists()


def test_validate_installation_nightly(mocker, install_mock, solc_binary, install_path):
    version = solcx.wrapper.get_solc_version(solc_binary)
    version_str_patch = mocker.patch("solcx.wrapper.get_version_str_from_solc_binary")
    version_str_patch.return_value = f"{version}-nightly"

    with pytest.warns(UnexpectedVersionWarning):
        solcx.install_solc()

    assert install_path.exists()


def test_validate_installation_fails(monkeypatch, solc_binary, install_path):
    def _mock(*args, **kwargs):
        if sys.platform == "win32":
            install_path.mkdir()
            with install_path.joinpath("solc.exe").open("w") as fp:
                fp.write("blahblah")
        else:
            with install_path.open("w") as fp:
                fp.write("blahblah")

    monkeypatch.setattr("solcx.install._install_solc_unix", _mock)
    monkeypatch.setattr("solcx.install._install_solc_windows", _mock)

    with pytest.raises(SolcInstallationError):
        solcx.install_solc()

    assert not install_path.exists()



import multiprocessing as mp
import threading

import pytest

import solcx


class ThreadWrap:
    def __init__(self, fn, *args, **kwargs):
        self.exc = None
        self.t = threading.Thread(target=self.wrap, args=(fn,) + args, kwargs=kwargs)
        self.t.start()

    def wrap(self, fn, *args, **kwargs):
        try:
            fn(*args, **kwargs)
        except Exception as e:
            self.exc = e

    def join(self):
        self.t.join()
        if self.exc is not None:
            raise self.exc


@pytest.mark.skipif("'--no-install' in sys.argv")
def test_threadlock(nosolc):
    threads = [ThreadWrap(solcx.install_solc, "0.6.9") for i in range(4)]
    for t in threads:
        t.join()


@pytest.mark.skipif("'--no-install' in sys.argv")
def test_processlock(nosolc):
    # have to use a context here to prevent a conflict with tqdm
    ctx = mp.get_context("spawn")
    threads = [ctx.Process(target=solcx.install_solc, args=("0.6.9",)) for i in range(4)]
    for t in threads:
        t.start()
    solcx.install_solc("0.6.9")
    for t in threads:
        t.join()
        assert t.exitcode == 0



import json
from pathlib import Path

import pytest

import solcx
from solcx.exceptions import ContractsNotFound

# these values should work for all compatible solc versions
combined_json_values = (
    "abi",
    "asm",
    "ast",
    "bin",
    "bin-runtime",
    "devdoc",
    "metadata",
    "opcodes",
    "srcmap",
    "srcmap-runtime",
    "userdoc",
)


@pytest.fixture(autouse=True)
def setup(all_versions):
    pass


def test_compile_source(foo_source):
    output = solcx.compile_source(foo_source)
    assert "<stdin>:Foo" in output
    for value in combined_json_values:
        assert value in output["<stdin>:Foo"]


def test_compile_source_with_optimization(foo_source):
    output = solcx.compile_source(foo_source, optimize=True, optimize_runs=10000)
    assert "<stdin>:Foo" in output
    for value in combined_json_values:
        assert value in output["<stdin>:Foo"]


def test_compile_source_empty():
    with pytest.raises(ContractsNotFound):
        solcx.compile_source("  ")


def test_compile_source_allow_empty():
    assert solcx.compile_source("  ", allow_empty=True) == {}


def test_compile_source_output_dir(tmp_path, foo_source):
    solcx.compile_source(foo_source, output_dir=tmp_path)
    output_path = tmp_path.joinpath("combined.json")

    with output_path.open() as fp:
        output = json.load(fp)["contracts"]
    assert "<stdin>:Foo" in output


def test_compile_source_overwrite(tmp_path, foo_source):
    output_path = tmp_path.joinpath("combined.json")
    with output_path.open("w") as fp:
        fp.write("foobar")

    with pytest.raises(FileExistsError):
        solcx.compile_source(foo_source, output_dir=tmp_path)
    with pytest.raises(FileExistsError):
        solcx.compile_source(foo_source, output_dir=tmp_path.joinpath("combined.json"))

    solcx.compile_source(foo_source, output_dir=tmp_path, overwrite=True)

    with output_path.open() as fp:
        output = json.load(fp)["contracts"]
    assert "<stdin>:Foo" in output


@pytest.mark.parametrize("key", combined_json_values)
def test_compile_source_output_types(foo_source, key):
    output = solcx.compile_source(foo_source, output_values=[key])
    assert list(output["<stdin>:Foo"]) == [key]


def test_compile_source_import_remapping(foo_path, bar_path, baz_source):
    import_remappings = {"contracts": foo_path.parent, "other": bar_path.parent}
    output = solcx.compile_source(baz_source, import_remappings=import_remappings)

    assert set(output) == {
        f"{foo_path.as_posix()}:Foo",
        f"{bar_path.as_posix()}:Bar",
        "<stdin>:Baz",
    }


def test_solc_binary(wrapper_mock, foo_source):
    wrapper_mock.expect(solc_binary=Path("path/to/solc"))
    solcx.compile_source(foo_source, ["abi"], solc_binary=Path("path/to/solc"), allow_empty=True)


def test_solc_version(wrapper_mock, all_versions, foo_source):
    solc_binary = solcx.install.get_executable(all_versions)
    wrapper_mock.expect(solc_binary=solc_binary)
    solcx.compile_source(foo_source, ["abi"], solc_version=all_versions, allow_empty=True)


def test_value_kwargs(compile_combined_json_mock, foo_source):
    compile_combined_json_mock.expect(random_kwarg="random-value")
    solcx.compile_source(foo_source, output_values=["abi"], random_kwarg="random-value")



import solcx


def test_import_solc(mocker, solc_binary, nosolc):
    version = solcx.wrapper.get_solc_version(solc_binary)
    patch = mocker.patch("solcx.install._get_which_solc")
    patch.return_value = solc_binary

    assert version in solcx.import_installed_solc()
    assert nosolc.joinpath(f"solc-v{version.base_version}").exists()


def test_import_solc_fails_after_importing(monkeypatch, solc_binary, nosolc):
    count = 0
    version = solcx.wrapper.get_solc_version(solc_binary)

    def version_mock(*args):
        # the first version call succeeds, the second attempt fails
        # this mocks a solc binary that no longer works after being copied
        nonlocal count
        if not count:
            count += 1
            return version
        raise Exception

    monkeypatch.setattr("solcx.install._get_which_solc", lambda: solc_binary)
    monkeypatch.setattr("solcx.wrapper.get_solc_version", version_mock)

    assert solcx.import_installed_solc() == []
    assert not nosolc.joinpath(solc_binary.name).exists()



import solcx
import solcx.main


def test_get_combined_json_outputs_defaults(mocker, foo_source):
    # verify we get the correct combined json outputs for different
    # compiler versions
    spy = mocker.spy(solcx.main, "_get_combined_json_outputs")

    solcx.compile_source(foo_source, solc_version="0.4.12")
    assert "function-debug" not in spy.spy_return

    solcx.compile_source(foo_source, solc_version="0.8.9")
    assert "function-debug" in spy.spy_return



from pathlib import Path

import pytest

import solcx

source = """pragma solidity >=0.4.11;
library UnlinkedLib {
    function linkMethod(uint _value, uint _multiplier) public returns (uint) {
        return _value * _multiplier;
    }
}

library OtherUnlinkedLib {
    function otherLinkMethod(uint _value, uint _multiplier) public returns (uint) {
        return _value * _multiplier;
    }
}

contract LinkTester {
    function testLibraryLinks(uint amount, uint multiple) external returns (uint) {
        uint a = UnlinkedLib.linkMethod(amount, multiple);
        return OtherUnlinkedLib.otherLinkMethod(a, multiple);
    }
}"""

addr1 = "0x1234567890123456789012345678901234567890"
addr2 = "0xdeadbeefdeadbeefdeadbeefdeadbeefdeadbeef"


@pytest.fixture
def bytecode(all_versions):
    yield solcx.compile_source(source)["<stdin>:LinkTester"]["bin"]


def test_unlinked_bytecode(bytecode):
    assert "_" in bytecode
    assert addr1[2:] not in bytecode
    assert addr2[2:] not in bytecode


def test_partial_link(bytecode):
    output = solcx.link_code(bytecode, {"<stdin>:UnlinkedLib": addr1})

    assert output != bytecode
    assert "_" in output
    assert addr1[2:] in output


def test_full_link(bytecode):
    output = solcx.link_code(
        bytecode, {"<stdin>:UnlinkedLib": addr1, "<stdin>:OtherUnlinkedLib": addr2}
    )

    # fully linked bytecode should be able to be interpreted as hex
    assert int(output, 16)

    assert addr1[2:] in output
    assert addr2[2:] in output


def test_solc_binary(wrapper_mock):
    wrapper_mock.expect(solc_binary=Path("path/to/solc"))
    solcx.link_code("0x00", {}, solc_binary=Path("path/to/solc"))


def test_solc_version(wrapper_mock, all_versions):
    solc_binary = solcx.install.get_executable(all_versions)
    wrapper_mock.expect(solc_binary=solc_binary)
    solcx.link_code("0x00", {}, solc_version=all_versions)



#!/usr/bin/python3

import pytest
from packaging.version import Version
from requests import ConnectionError

import solcx

_installed: dict = {}
VERSIONS = []


def pytest_addoption(parser):
    parser.addoption(
        "--no-install",
        action="store_true",
        help="Only run solcx tests against already installed solc versions",
    )
    parser.addoption(
        "--solc-versions",
        action="store",
        help="Only run tests against a specific version(s) of solc",
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "min_solc: minimum version of solc to run test against")


def pytest_collection(session):
    global VERSIONS
    if session.config.getoption("--solc-versions"):
        VERSIONS = [Version(i) for i in session.config.getoption("--solc-versions").split(",")]
    elif session.config.getoption("--no-install"):
        VERSIONS = solcx.get_installed_solc_versions()
    else:
        try:
            VERSIONS = solcx.get_installable_solc_versions()
        except ConnectionError:
            raise pytest.UsageError(
                "ConnectionError while attempting to get solc versions.\n"
                "Use the --no-install flag to only run tests against already installed versions."
            )
        for version in VERSIONS:
            solcx.install_solc(version)


# auto-parametrize the all_versions fixture with all target solc versions
def pytest_generate_tests(metafunc):
    if "all_versions" in metafunc.fixturenames:
        versions = VERSIONS.copy()
        for marker in metafunc.definition.iter_markers(name="min_solc"):
            versions = [i for i in versions if i >= Version(marker.args[0])]
        metafunc.parametrize("all_versions", versions, indirect=True)


@pytest.fixture
def all_versions(request):
    """
    Run a test against all solc versions.
    """
    version = request.param
    solcx.set_solc_version(version)
    return version


@pytest.fixture(scope="session")
def foo_source():
    yield """pragma solidity >=0.4.1;

contract Foo {
    function return13() public returns (uint a) {
        return 13;
    }
}
"""


@pytest.fixture(scope="session")
def bar_source():
    yield """
pragma solidity >=0.4.1;

import "../contracts/Foo.sol";

contract Bar is Foo {
    function getFunky() public returns (bytes4) {
        return 0x420Faded;
    }
}"""


@pytest.fixture(scope="session")
def baz_source():
    yield """
pragma solidity >=0.4.1;

import "../other/Bar.sol";

contract Baz is Bar {
    function doStuff() public returns (uint a) {
        return 31337;
    }
}"""


@pytest.fixture()
def invalid_source():
    yield """pragma solidity >=0.4.1;
contract Foo {"""


@pytest.fixture(scope="session")
def foo_path(tmp_path_factory, foo_source):
    source = tmp_path_factory.mktemp("contracts", False).joinpath("Foo.sol")
    with source.open("w") as fp:
        fp.write(foo_source)
    return source


@pytest.fixture(scope="session")
def bar_path(tmp_path_factory, bar_source):
    source = tmp_path_factory.mktemp("other", False).joinpath("Bar.sol")
    with source.open("w") as fp:
        fp.write(bar_source)
    return source


@pytest.fixture(scope="session")
def baz_path(tmp_path_factory, baz_source):
    source = tmp_path_factory.mktemp("baz", False).joinpath("Baz.sol")
    with source.open("w") as fp:
        fp.write(baz_source)
    return source



import re
import subprocess
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from packaging.version import Version

from solcx import install
from solcx.exceptions import SolcError, UnknownOption, UnknownValue

# (major.minor.patch)(nightly)(commit)
VERSION_REGEX = r"(\d+\.\d+\.\d+)(?:-nightly.\d+.\d+.\d+|)(\+commit.\w+)"


def get_version_str_from_solc_binary(solc_binary: Union[Path, str]) -> str:
    stdout_data = subprocess.check_output([str(solc_binary), "--version"], encoding="utf8")
    if not (match := next(re.finditer(VERSION_REGEX, stdout_data), None)):
        raise SolcError("Could not determine the solc binary version")

    # NOTE: May include "-nightly" suffix.
    return "".join(match.groups())


def get_solc_version(solc_binary: Union[Path, str], with_commit_hash: bool = False) -> Version:
    version_str = get_version_str_from_solc_binary(solc_binary)
    version = Version(version_str.replace("-nightly", ""))
    return version if with_commit_hash else Version(version.base_version)


def _to_string(key: str, value: Any) -> str:
    # convert data into a string prior to calling `solc`
    if isinstance(value, (int, str)):
        return str(value)
    elif isinstance(value, Path):
        return value.as_posix()
    elif isinstance(value, (list, tuple)):
        return ",".join(_to_string(key, i) for i in value)
    else:
        raise TypeError(f"Invalid type for {key}: {type(value)}")


def solc_wrapper(
    solc_binary: Optional[Union[Path, str]] = None,
    stdin: Optional[str] = None,
    source_files: Optional[Union[List, Path, str]] = None,
    import_remappings: Optional[Union[Dict, List, str]] = None,
    success_return_code: Optional[int] = None,
    **kwargs: Any,
) -> Tuple[str, str, List, subprocess.Popen]:
    """
    Wrapper function for calling to ``solc``.

    Args:
      solc_binary (Optional[Union[Path, str]]): Location of the
        ``solc`` binary. If not given, the current default binary is used.
      stdin (Optional[str]): Input to pass to ``solc`` via stdin.
      source_files (Optional[Union[List, Path, str]]): Path, or list of
        paths, of sources to compile
      import_remappings (Optional[Union[Dict, List, str]]): Path remappings.
        May be given as a string or list of strings, formatted as ``"prefix=path"``
        or a dict of ``{"prefix": "path"}``.
      success_return_code (Optional[int]): Expected exit code.
        Raises :class:`~solcx.exceptions.SolcError`` if the process returns a
        different value.
      **kwargs (Any): Flags to be passed to ``solc``. Keywords are converted to
        flags by prepending ``--`` and replacing ``_`` with ``-``, for example the
        keyword ``evm_version`` becomes ``--evm-version``. Values may be given in
        the following formats:

            * ``False``, ``None``: ignored
            * ``True``: flag is used without any arguments
            * str: given as an argument without modification
            * int: given as an argument, converted to a string
            * Path: converted to a string via ``Path.as_posix()``
            * List, Tuple: elements are converted to strings and joined with ``,``

    Returns:
      str: Process ``stdout`` output.
      str: Process ``stderr`` output.
      List: Full command executed by the function.
      Popen: Subprocess object used to call ``solc``.
    """
    solc_binary = Path(solc_binary) if solc_binary else install.get_executable()
    solc_version = get_solc_version(solc_binary)
    command: List = [str(solc_binary)]

    if (
        success_return_code is None
        and ("--help" in command or "help" in kwargs)
        and Version(solc_version.base_version) < Version("0.8.10")
    ):
        success_return_code = 1
    else:
        success_return_code = success_return_code or 0

    if source_files is not None:
        if isinstance(source_files, (str, Path)):
            command.append(_to_string("source_files", source_files))
        else:
            command.extend([_to_string("source_files", i) for i in source_files])

    if import_remappings is not None:
        if isinstance(import_remappings, str):
            command.append(import_remappings)
        else:
            if isinstance(import_remappings, dict):
                import_remappings = [f"{k}={v}" for k, v in import_remappings.items()]
            command.extend(import_remappings)

    for key, value in kwargs.items():
        if value is None or value is False:
            continue

        key = f"--{key.replace('_', '-')}"
        if value is True:
            command.append(key)
        else:
            command.extend([key, _to_string(key, value)])

    if "standard_json" not in kwargs and not source_files:
        # indicates that solc should read from stdin
        command.append("-")

    if stdin is not None:
        stdin = str(stdin)

    proc = subprocess.Popen(
        command,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf8",
    )

    stdoutdata, stderrdata = proc.communicate(stdin)

    stderrdata = (
        stderrdata.replace("Error: ", "") if stderrdata.startswith("Error: ") else stderrdata
    )

    if proc.returncode != success_return_code:
        if stderrdata.startswith("unrecognised option"):
            # unrecognised option '<FLAG>'
            flag = stderrdata.split("'")[1]
            raise UnknownOption(
                f"solc {solc_version.base_version} does not support the '{flag}' option'"
            )
        if stderrdata.startswith("Invalid option"):
            # Invalid option to <FLAG>: <OPTION>
            flag, option = stderrdata.split(": ")
            flag = flag.split(" ")[-1]
            raise UnknownValue(
                f"solc {solc_version.base_version} does not accept "
                f"'{option}' as an option for the '{flag}' flag"
            )

        raise SolcError(
            command=command,
            return_code=proc.returncode,
            stdin_data=stdin,
            stdout_data=stdoutdata,
            stderr_data=stderrdata,
        )

    return stdoutdata, stderrdata, command, proc



import os
import shutil
import subprocess
import sys
from pathlib import Path

import pytest
from packaging.version import Version

import solcx
from solcx.install import _download_solc


@pytest.fixture(scope="session")
def solc_binary():
    """
    Yields the path to the most recent solc binary.
    """
    version = solcx.get_installed_solc_versions()[0]
    yield solcx.install.get_executable(version)


@pytest.fixture
def nosolc(tmp_path, mocker):
    """
    Monkeypatch the install folder, so tests run with no versions of solc installed.

    Yields the temporary install folder.
    """
    install_patch = mocker.patch("solcx.install.get_solcx_install_folder")
    install_patch.return_value = tmp_path

    bin_patch = mocker.patch("solcx.install.get_default_solc_binary")
    bin_patch.return_value = None

    yield tmp_path


@pytest.fixture
def install_path(solc_binary, nosolc):
    """
    Yields the expected install path when using `install_mock`.
    """
    if sys.platform == "win32":
        yield nosolc.joinpath(solc_binary.parent.name)
    else:
        yield nosolc.joinpath(solc_binary.name)


@pytest.fixture
def install_mock(mocker, install_path, solc_binary):
    """
    Monkeypatches the install process by copying `solc_binary` to `install_path`.
    """

    def _mock(*args, **kwargs):
        if sys.platform == "win32":
            shutil.copytree(solc_binary.parent, install_path)
        else:
            shutil.copy(solc_binary, install_path)

    install_patch = mocker.patch("solcx.install._install_solc_unix")
    install_patch.side_effect = _mock

    windows_patch = mocker.patch("solcx.install._install_solc_windows")
    windows_patch.side_effect = _mock


class CompileMock:
    tarfile = None

    def __init__(self, solc_binary):
        self.solc_binary = solc_binary
        self.raise_cmd = None

    def __call__(self, cmd, **kwargs):
        if cmd[0] == self.raise_cmd:
            raise subprocess.CalledProcessError(1, cmd)
        if cmd == ["make"]:
            install_folder = Path(os.getcwd()).joinpath("solc")
            install_folder.mkdir()
            shutil.copy(self.solc_binary, install_folder.joinpath("solc"))

    def raise_on(self, cmd):
        self.raise_cmd = cmd

    @classmethod
    def download(cls, url, show_progress):
        if not cls.tarfile:
            cls.tarfile = _download_solc(url, show_progress)

        return cls.tarfile


@pytest.fixture
def compile_mock(monkeypatch, nosolc, solc_binary):
    """
    Monkeypatches `compile_solc`.

    * The first use downloads a tarfile, which is then returned for all subsequent uses
    * The call to `make` copies `solc_binary` to the expected location
    * `compile_make.raise_on` tells the mock to raise `CalledProcessError` on a specific command
    """
    mock = CompileMock(solc_binary)
    monkeypatch.setattr("subprocess.check_call", mock)
    monkeypatch.setattr("solcx.install._download_solc", mock.download)
    yield mock


@pytest.fixture
def cwd():
    """
    Yields the current working directory, and ensures it is reset during teardown.
    """
    cwd = os.getcwd()
    yield cwd
    os.chdir(cwd)


@pytest.fixture
def pragmapatch(monkeypatch):
    """
    Monkeypatches the installed/available versions to ensure consistent testing of
    `install_solc_pragma` and `set_pragma`
    """
    monkeypatch.setattr(
        "solcx.install.get_installed_solc_versions",
        lambda: [
            Version("0.4.2"),
            Version("0.4.11"),
            Version("0.4.25"),
            Version("0.5.0"),
            Version("0.5.4"),
            Version("0.5.7"),
            Version("1.2.3"),
        ],
    )
    monkeypatch.setattr(
        "solcx.install.get_installable_solc_versions",
        lambda: [
            Version("0.4.11"),
            Version("0.4.24"),
            Version("0.4.26"),
            Version("0.5.3"),
            Version("0.6.0"),
        ],
    )
    monkeypatch.setattr("solcx.install.set_solc_version", lambda *args: None)



from pathlib import Path

import pytest

import solcx
from solcx.exceptions import ContractsNotFound, SolcError


@pytest.fixture
def input_json(all_versions):
    json = {
        "language": "Solidity",
        "sources": {},
        "settings": {"outputSelection": {"*": {"*": ["evm.bytecode.object"]}}},
    }
    yield json


def _compile_assertions(output_json, *contract_names):
    assert isinstance(output_json, dict)
    assert "contracts" in output_json
    contracts = output_json["contracts"]
    for key in contract_names:
        assert int(contracts[f"contracts/{key}.sol"][key]["evm"]["bytecode"]["object"], 16)


def test_compile_standard(input_json, foo_source):
    input_json["sources"] = {"contracts/Foo.sol": {"content": foo_source}}
    result = solcx.compile_standard(input_json)

    _compile_assertions(result, "Foo")


def test_compile_standard_invalid_source(input_json, invalid_source):
    input_json["sources"] = {"contracts/Foo.sol": {"content": invalid_source}}
    with pytest.raises(SolcError):
        solcx.compile_standard(input_json)


def test_compile_standard_with_dependency(input_json, foo_source, bar_source):
    input_json["sources"] = {
        "contracts/Foo.sol": {"content": foo_source},
        "contracts/Bar.sol": {"content": bar_source},
    }
    result = solcx.compile_standard(input_json)

    _compile_assertions(result, "Foo", "Bar")


def test_compile_standard_with_file_paths(input_json, foo_path):
    input_json["sources"] = {"contracts/Foo.sol": {"urls": [str(foo_path)]}}
    result = solcx.compile_standard(input_json, allow_paths=[foo_path.parent])

    _compile_assertions(result, "Foo")


def test_compile_standard_empty():
    with pytest.raises(ContractsNotFound):
        solcx.compile_standard({"language": "Solidity", "sources": {}})


def test_solc_binary(wrapper_mock, foo_source):
    wrapper_mock.expect(solc_binary=Path("path/to/solc"))
    solcx.compile_standard({}, solc_binary=Path("path/to/solc"), allow_empty=True)


def test_solc_version(wrapper_mock, all_versions, foo_source):
    solc_binary = solcx.install.get_executable(all_versions)
    wrapper_mock.expect(solc_binary=solc_binary)
    solcx.compile_standard({}, solc_version=all_versions, allow_empty=True)



import json
from pathlib import Path

import pytest

import solcx
from solcx.exceptions import ContractsNotFound, SolcError

# these values should work for all compatible solc versions
combined_json_values = (
    "abi",
    "asm",
    "ast",
    "bin",
    "bin-runtime",
    "devdoc",
    "metadata",
    "opcodes",
    "srcmap",
    "srcmap-runtime",
    "userdoc",
)


@pytest.fixture(autouse=True)
def setup(all_versions):
    pass


@pytest.fixture(scope="session")
def empty_path(tmp_path_factory, baz_source):
    source = tmp_path_factory.mktemp("empty", False).joinpath("Empty.sol")
    with source.open("w") as fp:
        fp.write(" ")
    return source


def test_compile_single_file(foo_path):
    output = solcx.compile_files([foo_path])
    assert f"{foo_path.as_posix()}:Foo" in output
    for value in combined_json_values:
        assert value in output[f"{foo_path.as_posix()}:Foo"]


def test_compile_single_file_no_sequence(foo_path):
    output = solcx.compile_files(foo_path)
    assert f"{foo_path.as_posix()}:Foo" in output
    for value in combined_json_values:
        assert value in output[f"{foo_path.as_posix()}:Foo"]


def test_compile_multiple_files(foo_path, bar_path, baz_path):
    output = solcx.compile_files([foo_path, bar_path, baz_path])

    assert set(output) == {
        f"{foo_path.as_posix()}:Foo",
        f"{bar_path.as_posix()}:Bar",
        f"{baz_path.as_posix()}:Baz",
    }


def test_compile_multiple_files_fails_bad_path(baz_path):
    with pytest.raises(SolcError):
        solcx.compile_files([baz_path])


def test_import_remappings(foo_path, bar_path, baz_path):
    import_remappings = {"contracts": foo_path.parent, "other": bar_path.parent}
    output = solcx.compile_files([baz_path], import_remappings=import_remappings)

    assert set(output) == {
        f"{foo_path.as_posix()}:Foo",
        f"{bar_path.as_posix()}:Bar",
        f"{baz_path.as_posix()}:Baz",
    }


def test_allow_paths(foo_path, bar_path, baz_path):
    output = solcx.compile_files([baz_path], allow_paths=[foo_path.parent.parent])

    assert set(output) == {
        f"{foo_path.as_posix()}:Foo",
        f"{bar_path.as_posix()}:Bar",
        f"{baz_path.as_posix()}:Baz",
    }


def test_compile_files_empty(empty_path):
    with pytest.raises(ContractsNotFound):
        solcx.compile_files([empty_path])


def test_compile_files_allow_empty(empty_path):
    assert solcx.compile_files([empty_path], allow_empty=True) == {}


@pytest.mark.parametrize("key", combined_json_values)
def test_compile_files_output_types(foo_path, key):
    output = solcx.compile_files([foo_path], output_values=[key])
    assert list(output[f"{foo_path.as_posix()}:Foo"]) == [key]


def test_compile_source_output_dir(tmp_path, foo_path, bar_path):
    solcx.compile_files([foo_path, bar_path], output_dir=tmp_path)
    output_path = tmp_path.joinpath("combined.json")

    with output_path.open() as fp:
        output = json.load(fp)["contracts"]

    assert f"{foo_path.as_posix()}:Foo" in output
    assert f"{bar_path.as_posix()}:Bar" in output


def test_solc_binary(wrapper_mock, foo_path):
    wrapper_mock.expect(solc_binary=Path("path/to/solc"))
    solcx.compile_files([foo_path], ["abi"], solc_binary=Path("path/to/solc"), allow_empty=True)


def test_solc_version(wrapper_mock, all_versions, foo_path):
    solc_binary = solcx.install.get_executable(all_versions)
    wrapper_mock.expect(solc_binary=solc_binary)
    solcx.compile_files([foo_path], ["abi"], solc_version=all_versions, allow_empty=True)


def test_value_kwargs(compile_combined_json_mock, foo_path):
    compile_combined_json_mock.expect(random_kwarg="random-value")
    solcx.compile_files(foo_path, output_values=["abi"], random_kwarg="random-value")



import warnings

import pytest

import solcx
from solcx.exceptions import SolcInstallationError


@pytest.mark.skipif("'--no-install' in sys.argv")
def test_install_latest():
    version = solcx.get_installable_solc_versions()[0]
    assert solcx.install_solc("latest") == version


def test_unknown_platform(monkeypatch):
    monkeypatch.setattr("sys.platform", "potatoOS")
    with pytest.raises(OSError):
        solcx.install_solc()


def test_install_unknown_version():
    with pytest.raises(SolcInstallationError):
        solcx.install_solc("0.4.99")


@pytest.mark.skipif("'--no-install' in sys.argv")
def test_progress_bar(nosolc):
    # There should be no warnings!
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        assert solcx.install_solc("0.6.9", show_progress=True).base_version == "0.6.9"


def test_environment_var_path(monkeypatch, tmp_path):
    install_folder = solcx.get_solcx_install_folder()
    monkeypatch.setenv("SOLCX_BINARY_PATH", tmp_path.as_posix())
    assert solcx.get_solcx_install_folder() != install_folder

    monkeypatch.undo()
    assert solcx.get_solcx_install_folder() == install_folder


def test_environment_var_versions(monkeypatch, tmp_path):
    versions = solcx.get_installed_solc_versions()
    monkeypatch.setenv("SOLCX_BINARY_PATH", tmp_path.as_posix())
    assert solcx.get_installed_solc_versions() != versions

    monkeypatch.undo()
    assert solcx.get_installed_solc_versions() == versions


@pytest.mark.skipif("'--no-install' in sys.argv")
def test_environment_var_install(monkeypatch, tmp_path):
    assert not tmp_path.joinpath("solc-v0.6.9").exists()

    monkeypatch.setenv("SOLCX_BINARY_PATH", tmp_path.as_posix())

    solcx.install_solc("0.6.9")
    assert tmp_path.joinpath("solc-v0.6.9").exists()



import pytest

import solcx
from solcx.exceptions import SolcNotInstalled, UnsupportedVersionError


def test_get_executable():
    assert solcx.install.get_executable() == solcx.install._default_solc_binary


def test_no_default_set(nosolc):
    with pytest.raises(SolcNotInstalled):
        solcx.install.get_executable()


def test_unsupported_version():
    with pytest.raises(UnsupportedVersionError):
        solcx.install.get_executable("0.4.0")


def test_version_not_installed():
    with pytest.raises(SolcNotInstalled):
        solcx.install.get_executable("999.999.999")



"""
Install solc
"""
import argparse
import logging
import os
import re
import shutil
import stat
import subprocess
import sys
import tarfile
import tempfile
import time
import warnings
import zipfile
from base64 import b64encode
from io import BytesIO
from pathlib import Path
from typing import Dict, List, Optional, Union

import requests
from packaging.specifiers import SpecifierSet
from packaging.version import InvalidVersion, Version

from solcx import wrapper
from solcx.exceptions import (
    DownloadError,
    SolcInstallationError,
    SolcNotInstalled,
    UnexpectedVersionError,
    UnexpectedVersionWarning,
    UnsupportedVersionError,
)
from solcx.utils.lock import get_process_lock

try:
    from tqdm import tqdm  # type: ignore[import-untyped]
except ImportError:
    tqdm = None


BINARY_DOWNLOAD_BASE = "https://solc-bin.ethereum.org/{}-amd64/{}"
SOURCE_DOWNLOAD_BASE = "https://github.com/ethereum/solidity/releases/download/v{}/{}"
GITHUB_RELEASES = "https://api.github.com/repos/ethereum/solidity/releases?per_page=100"

MINIMAL_SOLC_VERSION = Version("0.4.11")
LOGGER = logging.getLogger("solcx")

SOLCX_BINARY_PATH_VARIABLE = "SOLCX_BINARY_PATH"

_default_solc_binary = None


def _get_os_name() -> str:
    if sys.platform.startswith("linux"):
        return "linux"
    if sys.platform == "darwin":
        return "macosx"
    if sys.platform == "win32":
        return "windows"
    raise OSError(f"Unsupported OS: '{sys.platform}' - py-solc-x supports Linux, OSX and Windows")


def _convert_and_validate_version(version: Union[str, Version]) -> Version:
    # take a user-supplied version as a string or Version
    # validate the value, and return a Version object
    if not isinstance(version, Version):
        version = Version(version.lstrip("v"))
    if version not in SpecifierSet(">=0.4.11"):
        raise UnsupportedVersionError("py-solc-x does not support solc versions <0.4.11")
    return version


def _unlink_solc(solc_path: Path) -> None:
    solc_path.unlink()
    if _get_os_name() == "windows":
        shutil.rmtree(solc_path.parent)


def get_solcx_install_folder(solcx_binary_path: Optional[Union[Path, str]] = None) -> Path:
    """
    Return the directory where ``py-solc-x`` stores installed ``solc`` binaries.

    By default, this is ``~/.solcx``.

    Args:
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path, used
        to override the default installation directory.

    Returns:
      Path: Subdirectory where ``solc`` binaries are are saved.
    """
    if os.getenv(SOLCX_BINARY_PATH_VARIABLE):
        return Path(os.environ[SOLCX_BINARY_PATH_VARIABLE])
    elif solcx_binary_path is not None:
        return Path(solcx_binary_path)
    else:
        path = Path.home().joinpath(".solcx")
        path.mkdir(exist_ok=True)
        return path


def _get_which_solc() -> Optional[Path]:
    # get the path for the currently installed `solc` version, if any
    path = shutil.which("solc")
    return Path(path) if path else None


def import_installed_solc(solcx_binary_path: Optional[Union[Path, str]] = None) -> List[Version]:
    """
    Search for and copy installed ``solc`` versions into the local installation
    folder.

    Args:
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path, used to
        override the default installation directory.

    Returns:
      List: Imported solc versions
    """
    path_list = []
    if solc_in_path := _get_which_solc():
        path_list.append(solc_in_path)

    # on OSX, also copy all versions of solc from cellar
    if _get_os_name() == "macosx":
        path_list.extend(Path("/usr/local/Cellar").glob("solidity*/**/solc"))

    imported_versions = []
    solcx_binary_path = get_solcx_install_folder(solcx_binary_path=solcx_binary_path)
    installed_versions = get_installed_solc_versions(solcx_binary_path=solcx_binary_path)

    for path in path_list:
        try:
            version = wrapper.get_solc_version(path)
        except Exception:
            continue

        if version in installed_versions:
            # Already installed.
            continue

        base_path = get_solcx_install_folder(solcx_binary_path)
        copy_path = base_path.joinpath(f"solc-v{version.base_version}")
        if _get_os_name() == "windows":
            copy_path.mkdir()
            copy_path = copy_path.joinpath("solc.exe")

        shutil.copy(path, copy_path)

        try:
            version_check = wrapper.get_solc_version(copy_path)
        except Exception:
            _unlink_solc(copy_path)

        else:
            if version != version_check:
                _unlink_solc(copy_path)
            else:
                imported_versions.append(version)

    return imported_versions


def get_executable(
    version: Optional[Union[str, Version]] = None,
    solcx_binary_path: Optional[Union[Path, str]] = None,
) -> Path:
    """
    Return the Path to an installed `solc` binary.

    Args:
      version (Optional[Union[str, Version]]): Installed ``solc`` version
        to get the path of. If not given, returns the path of the active version.
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path, used to
        override the default installation directory.

    Returns:
      Path: ``solc`` executable.
    """
    if not version:
        if binary := get_default_solc_binary():
            return binary

        raise SolcNotInstalled(
            "Solc is not installed. Call solcx.get_installable_solc_versions()"
            " to view for available versions and solcx.install_solc() to install."
        )

    version = _convert_and_validate_version(version)
    solc_bin = get_solcx_install_folder(solcx_binary_path).joinpath(f"solc-v{version.base_version}")
    if _get_os_name() == "windows":
        solc_bin = solc_bin.joinpath("solc.exe")
    if not solc_bin.exists():
        raise SolcNotInstalled(
            f"solc {version.base_version} has not been installed."
            f" Use solcx.install_solc('{version.base_version}') to install."
        )
    return solc_bin


def get_default_solc_binary():
    return _default_solc_binary


def set_solc_version(
    version: Union[str, Version],
    silent: bool = False,
    solcx_binary_path: Optional[Union[Path, str]] = None,
) -> None:
    """
    Set the currently active ``solc`` binary.

    Args:
      version (Union[str, Version]): Installed ``solc`` version
        to get the path of. If not given, returns the path of the
        active version.
      silent (bool): If ``True``, do not generate any logger output.
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path,
        used to override the default installation directory.
    """
    version = _convert_and_validate_version(version)
    global _default_solc_binary
    _default_solc_binary = get_executable(version, solcx_binary_path)
    if not silent:
        LOGGER.info(f"Using solc version {version}")


def select_pragma_version(pragma_string: str, version_list: List[Version]) -> Optional[Version]:
    """
    Get a matching version from the given pragma string and a version list.

    Args:
        pragma_string (str): A pragma str.
        version_list (List[Version]): A list of valid versions.

    Returns:
        Optional[Version]: A selected version from the given list.
    """

    comparator_set_range = pragma_string.replace(" ", "").split("||")
    comparator_regex = re.compile(r"(([<>]?=?|\^)\d+\.\d+\.\d+)")
    version = None

    def _as_spec(item: str) -> str:
        ret = item.replace("^", "~=")

        if ret and ret[0].isnumeric():
            return f"=={ret}"

        elif ret and len(ret) >= 2 and ret[0] == "=" and ret[1] != "=":
            return f"={ret}"

        return ret

    for comparator_set in comparator_set_range:
        specs = ",".join([_as_spec(i[0]) for i in comparator_regex.findall(comparator_set)])
        spec = SpecifierSet(specs)
        matching = sorted(list(spec.filter(version_list)), reverse=True)
        selected = matching[0] if matching else None
        if selected and (not version or version < selected):
            version = selected

    return version


def set_solc_version_pragma(
    pragma_string: str, silent: bool = False, check_new: bool = False
) -> Version:
    """
    Set the currently active ``solc`` binary based on a pragma statement.

    The newest installed version that matches the pragma is chosen. =

    Raises:
      :class:`~solcx.exceptions.SolcNotInstalled`: When no installed versions match.

    Args:
      pragma_string (str): Pragma statement, e.g. ``"pragma solidity ^0.4.22;"``
      silent (bool): If ``True``, do not generate any logger output. Defaults to
        ``False``.
      check_new (bool): If ``True``, also check if there is a newer compatible
        version that has not been installed. Defaults to ``False``.

    Returns:
      Version: The new active `solc` version.
    """
    installed_versions = get_installed_solc_versions()
    if not (version := select_pragma_version(pragma_string, installed_versions)):
        raise SolcNotInstalled(
            f"No compatible solc version installed."
            f" Use solcx.install_solc_version_pragma('{pragma_string}') to install."
        )
    set_solc_version(version, silent)
    if check_new:
        latest = install_solc_pragma(pragma_string, False)
        if latest > version:
            LOGGER.info(f"Newer compatible solc version exists: {latest}")

    return version


def install_solc_pragma(
    pragma_string: str,
    install: bool = True,
    show_progress: bool = False,
    solcx_binary_path: Optional[Union[Path, str]] = None,
) -> Version:
    """
    Find, and optionally install, the latest compatible ``solc`` version
    based on a pragma statement.

    Args:
      pragma_string (str):  Pragma statement, e.g.
      ``"pragma solidity ^0.4.22;"``.
      install (bool): If ``True``, installs the version of ``solc``. Defaults
        to ``True``.
      show_progress (bool): If ``True``, display a progress bar while
        downloading. Requires installing the ``tqdm`` package. Defaults to
        ``False``.
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path,
        used to override the default installation directory.

    Returns:
      Version: Installed ``solc`` version.
    """

    installed_versions = get_installable_solc_versions()
    if version := select_pragma_version(pragma_string, installed_versions):
        install_solc(version, show_progress=show_progress, solcx_binary_path=solcx_binary_path)
    else:
        raise UnsupportedVersionError(
            f"Compatible solc version does not exist matching '{pragma_string}'."
        )

    return version


def get_installable_solc_versions() -> List[Version]:
    """
    Return a list of all ``solc`` versions that can be installed by py-solc-x.

    Returns:
      List: List of Versions objects of installable ``solc`` versions.
    """
    data = requests.get(BINARY_DOWNLOAD_BASE.format(_get_os_name(), "list.json"))
    if data.status_code != 200:
        raise ConnectionError(
            f"Status {data.status_code} when getting solc versions from solc-bin.ethereum.org"
        )
    version_list = sorted((Version(i) for i in data.json()["releases"]), reverse=True)
    version_list = [i for i in version_list if i >= MINIMAL_SOLC_VERSION]
    return version_list


def get_compilable_solc_versions(
    headers: Optional[Dict] = None, rate_limit_wait_time: float = 3.0
) -> List[Version]:
    """
    Return a list of all ``solc`` versions that can be compiled from source by py-solc-x.

    Args:
      headers (Optional[Dict]): Headers to include in the request to Github.
      rate_limit_wait_time (float): A value to use when waiting for rate-limiting.
        Defaults to 3.0 but will increment to the value plus it's half for the
        subsequent call.

    Returns:
      List: List of Versions objects of installable `solc` versions.
    """
    if _get_os_name() == "windows":
        raise OSError("Compiling from source is not supported on Windows systems")

    version_list = []
    pattern = "solidity_[0-9].[0-9].[0-9]{1,}.tar.gz"

    if headers is None and os.getenv("GITHUB_TOKEN") is not None:
        auth = b64encode(os.environ["GITHUB_TOKEN"].encode()).decode()
        headers = {"Authorization": f"Basic {auth}"}

    data = requests.get(GITHUB_RELEASES, headers=headers)
    if data.status_code != 200:
        msg = (
            f"Status {data.status_code} when getting solc versions from Github:"
            f" '{data.json()['message']}'"
        )
        if data.status_code == 403:
            msg += (
                "\n\nIf this issue persists, generate a Github API token and store"
                " it as the environment variable `GITHUB_TOKEN`:\n"
                "https://github.blog/2013-05-16-personal-api-tokens/"
            )
            LOGGER.warning(msg)

            # Wait and retry
            time.sleep(rate_limit_wait_time)
            return get_compilable_solc_versions(
                headers, rate_limit_wait_time=rate_limit_wait_time + (rate_limit_wait_time / 2)
            )

        raise ConnectionError(msg)

    for release in data.json():
        try:
            version = Version(release["tag_name"].lstrip("v"))
        except InvalidVersion:
            # ignore non-standard releases (e.g. the 0.8.x preview)
            continue

        asset = next((i for i in release["assets"] if re.match(pattern, i["name"])), False)
        if asset:
            version_list.append(version)
        if version == MINIMAL_SOLC_VERSION:
            break
    return sorted(version_list, reverse=True)


def get_installed_solc_versions(
    solcx_binary_path: Optional[Union[Path, str]] = None
) -> List[Version]:
    """
    Return a list of currently installed ``solc`` versions.

    Args:
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path,
        used to override the default installation directory.

    Returns:
      List: List of Version objects of installed ``solc`` versions.
    """
    install_path = get_solcx_install_folder(solcx_binary_path=solcx_binary_path)
    return sorted([Version(i.name[6:]) for i in install_path.glob("solc-v*")], reverse=True)


def install_solc(
    version: Union[str, Version] = "latest",
    show_progress: bool = False,
    solcx_binary_path: Optional[Union[Path, str]] = None,
) -> Version:
    """
    Download and install a precompiled version of ``solc``.

    Args:
      version (Union[str, Version]): Version of ``solc`` to install.
        Default is the newest available version.
      show_progress (bool): If ``True``, display a progress bar while
        downloading. Requires installing the ``tqdm`` package.
      solcx_binary_path (Optional[Union[Path, str]]): User-defined path,
        used to override the default installation directory.

    Returns:
      Version: installed solc version
    """

    version = (
        get_installable_solc_versions()[0]
        if version == "latest"
        else _convert_and_validate_version(version)
    )

    os_name = _get_os_name()
    process_lock = get_process_lock(str(version))

    with process_lock:
        if _check_for_installed_version(version, solcx_binary_path):
            path = get_solcx_install_folder(solcx_binary_path=solcx_binary_path).joinpath(
                f"solc-v{version}"
            )
            LOGGER.info(f"solc {version} already installed at: {path}")
            return version

        data = requests.get(BINARY_DOWNLOAD_BASE.format(_get_os_name(), "list.json"))
        if data.status_code != 200:
            raise ConnectionError(
                f"Status {data.status_code} when getting solc versions from solc-bin.ethereum.org"
            )
        try:
            filename = data.json()["releases"][str(version)]
        except KeyError:
            raise SolcInstallationError(f"Solc binary for v{version} is not available for this OS")

        if os_name in ("linux", "macosx"):
            _install_solc_unix(
                version, filename, show_progress, solcx_binary_path=solcx_binary_path
            )
        elif os_name == "windows":
            _install_solc_windows(
                version, filename, show_progress, solcx_binary_path=solcx_binary_path
            )

        base_version = version if isinstance(version, str) else version.base_version
        try:
            _validate_installation(Version(base_version), solcx_binary_path=solcx_binary_path)
        except SolcInstallationError as exc:
            if os_name != "windows":
                exc.args = (
                    f"{exc.args[0]} If this issue persists, you can try to compile from "
                    f"source code using `solcx.compile_solc('{version}')`.",
                )
            raise exc

    return version


def compile_solc(
    version: Optional[Union[str, Version]] = None,
    show_progress: bool = False,
    solcx_binary_path: Optional[Union[Path, str]] = None,
) -> Version:
    """
    Install a version of `solc` by downloading and compiling source code.

    Args:
        version (Union[str, Version]): Version of `solc` to install.
          Use literal ``"latest"`` or ``None`` to install the latest version.
          Default is the latest.
        show_progress (bool): If ``True``, display a progress bar while downloading.
          Requires installing the `tqdm` package.
        solcx_binary_path (Optional[Union[Path, str]]): User-defined path, used to
          override the default installation directory.

    Returns:
        Version: The version of the compiler installed.
    """
    if _get_os_name() == "windows":
        raise OSError("Compiling from source is not supported on Windows systems")

    if version in (None, "latest"):
        solc_version = get_compilable_solc_versions()[0]

    elif version is not None:
        solc_version = _convert_and_validate_version(version)

    else:
        # Shouldn't get here - but is for type-check.
        raise ValueError(version)

    process_lock = get_process_lock(str(solc_version))

    with process_lock:
        if _check_for_installed_version(solc_version, solcx_binary_path):
            path = get_solcx_install_folder(solcx_binary_path).joinpath(
                f"solc-v{solc_version.base_version}"
            )
            LOGGER.info(f"solc {solc_version.base_version} already installed at: {path}")
            return solc_version

        temp_path = _get_temp_folder()
        download = SOURCE_DOWNLOAD_BASE.format(
            solc_version.base_version, f"solidity_{solc_version.base_version}.tar.gz"
        )
        install_path = get_solcx_install_folder(solcx_binary_path).joinpath(
            f"solc-v{solc_version.base_version}"
        )

        content = _download_solc(download, show_progress)
        with tarfile.open(fileobj=BytesIO(content)) as tar:
            tar.extractall(temp_path)

        temp_path = temp_path / f"solidity_{solc_version.base_version}"

        LOGGER.info("Running dependency installation script `install_deps.sh`...")
        install_script = temp_path / "scripts" / "install_deps.sh"
        if install_script.is_file():
            try:
                subprocess.check_call(["sh", install_script], stderr=subprocess.DEVNULL)
            except subprocess.CalledProcessError as exc:
                LOGGER.warning(exc, exc_info=True)

        original_path = os.getcwd()
        build_path = temp_path / "build"
        build_path.mkdir(parents=True, exist_ok=True)
        os.chdir(str(build_path.resolve()))

        try:
            for cmd in (["cmake", ".."], ["make"]):
                LOGGER.info(f"Running `{cmd[0]}`...")
                subprocess.check_call(cmd, stderr=subprocess.DEVNULL)

        except subprocess.CalledProcessError as exc:
            err_msg = (
                f"{cmd[0]} returned non-zero exit status {exc.returncode}"
                " while attempting to build solc from the source.\n"
                "This is likely due to a missing or incorrect version of a build dependency."
            )
            if _get_os_name() == "macosx":
                err_msg = (
                    f"{err_msg}\n\nFor suggested installation options: "
                    "https://github.com/iamdefinitelyahuman/py-solc-x/wiki/Installing-Solidity-on-OSX"  # noqa: E501
                )
            raise SolcInstallationError(err_msg)

        finally:
            os.chdir(original_path)

        solc_path = build_path / "solc" / "solc"
        solc_path.rename(install_path)
        install_path.chmod(install_path.stat().st_mode | stat.S_IEXEC)
        _validate_installation(solc_version, solcx_binary_path)

    return solc_version


def _check_for_installed_version(
    version: Union[str, Version], solcx_binary_path: Optional[Union[Path, str]] = None
) -> bool:
    path = get_solcx_install_folder(solcx_binary_path).joinpath(f"solc-v{version}")
    return path.exists()


def _get_temp_folder() -> Path:
    path = Path(tempfile.gettempdir()).joinpath(f"solcx-tmp-{os.getpid()}")
    if path.exists():
        shutil.rmtree(str(path))
    path.mkdir()
    return path


def _download_solc(url: str, show_progress: bool, rate_limit_wait_time: float = 3.0) -> bytes:
    LOGGER.info(f"Downloading from {url}")
    response = requests.get(url, stream=show_progress)
    if response.status_code == 404:
        raise DownloadError(
            "404 error when attempting to download from {} - are you sure this"
            " version of solidity is available?".format(url)
        )

    elif response.status_code == 403 and "API rate limit exceeded" in response.text:
        # Handle GitHub API rate limiting.
        time_to_wait = rate_limit_wait_time or 1  # Prevent accidents
        time.sleep(time_to_wait)
        return _download_solc(url, show_progress, rate_limit_wait_time + (rate_limit_wait_time / 2))

    elif response.status_code != 200:
        raise DownloadError(
            f"Received status code {response.status_code} when attempting to download from {url}"
        )

    if show_progress and not tqdm:
        LOGGER.warning("Must install `tqdm` to see download progress")
        return response.content

    elif not show_progress:
        return response.content

    total_size = int(response.headers.get("content-length", 0))
    progress_bar = tqdm(total=total_size, unit="iB", unit_scale=True)
    content = bytes()

    for data in response.iter_content(1024, decode_unicode=True):
        progress_bar.update(len(data))
        content += data
    progress_bar.close()

    return content


def _install_solc_unix(
    version: Version, filename: str, show_progress: bool, solcx_binary_path: Union[Path, str, None]
) -> None:
    download = BINARY_DOWNLOAD_BASE.format(_get_os_name(), filename)
    install_path = get_solcx_install_folder(solcx_binary_path=solcx_binary_path).joinpath(
        f"solc-v{version}"
    )

    content = _download_solc(download, show_progress)
    with open(install_path, "wb") as fp:
        fp.write(content)

    install_path.chmod(install_path.stat().st_mode | stat.S_IEXEC)


def _install_solc_windows(
    version: Version, filename: str, show_progress: bool, solcx_binary_path: Union[Path, str, None]
) -> None:
    download = BINARY_DOWNLOAD_BASE.format(_get_os_name(), filename)
    install_path = get_solcx_install_folder(solcx_binary_path).joinpath(f"solc-v{version}")

    temp_path = _get_temp_folder()
    content = _download_solc(download, show_progress)

    if Path(filename).suffix == ".exe":
        install_path.mkdir()
        with open(install_path.joinpath("solc.exe"), "wb") as fp:
            fp.write(content)

    else:
        with zipfile.ZipFile(BytesIO(content)) as zf:
            zf.extractall(str(temp_path))
            shutil.move(temp_path, install_path)


def _validate_installation(version: Version, solcx_binary_path: Union[Path, str, None]) -> None:
    binary_path = get_executable(version, solcx_binary_path)
    try:
        installed_version = wrapper.get_version_str_from_solc_binary(binary_path)
    except Exception:
        _unlink_solc(binary_path)
        raise SolcInstallationError(
            "Downloaded binary would not execute, or returned unexpected output."
        )

    installed_version_clean = Version(
        Version(installed_version.replace("-nightly", "")).base_version
    )
    base_version = version if isinstance(version, str) else version.base_version
    if installed_version_clean.base_version != base_version:
        # Without the nightly suffix, it should be the same!
        _unlink_solc(binary_path)
        raise UnexpectedVersionError(
            f"Attempted to install solc v{version}, but got solc v{installed_version}"
        )
    if installed_version_clean not in (
        Version(base_version),
        f"{base_version}",
    ) or installed_version.endswith("-nightly"):
        # If it does have the nightly suffix, then only warn.
        warnings.warn(
            f"Installed solc version is v{installed_version_clean}, expecting v{base_version}",
            UnexpectedVersionWarning,
        )

    if not get_default_solc_binary():
        set_solc_version(version)

    LOGGER.info(f"solc {version} successfully installed at: {binary_path}")


try:
    # try to set the result of `which`/`where` as the default
    if path := _get_which_solc():
        _default_solc_binary = path
except Exception:
    # if not available, use the most recent solcx installed version
    if get_installed_solc_versions():
        set_solc_version(get_installed_solc_versions()[0], silent=True)


if __name__ == "__main__":
    argument_parser = argparse.ArgumentParser()
    argument_parser.add_argument("version")
    argument_parser.add_argument("--solcx-binary-path", default=None)
    args = argument_parser.parse_args()
    install_solc(args.version, solcx_binary_path=args.solcx_binary_path)



import os

import pytest

import solcx
from solcx.exceptions import SolcInstallationError


@pytest.mark.skipif("sys.platform != 'win32'")
def test_fails_on_windows():
    with pytest.raises(OSError):
        solcx.compile_solc("latest")


@pytest.mark.skipif("sys.platform == 'win32'")
def test_compile_already_installed():
    version = solcx.get_installed_solc_versions()[0]
    assert solcx.compile_solc("latest") == version


@pytest.mark.skipif("sys.platform == 'win32'")
def test_compile(compile_mock, solc_binary, cwd):
    version = solcx.wrapper.get_solc_version(solc_binary)
    solcx.compile_solc(version)

    assert os.getcwd() == cwd
    assert solcx.get_installed_solc_versions() == [version]


@pytest.mark.skipif("sys.platform == 'win32'")
def test_compile_install_deps_fails(compile_mock, solc_binary, cwd):
    version = solcx.wrapper.get_solc_version(solc_binary)
    compile_mock.raise_on("sh")
    solcx.compile_solc(version)

    assert os.getcwd() == cwd
    assert solcx.get_installed_solc_versions() == [version]


@pytest.mark.skipif("sys.platform == 'win32'")
def test_compile_install_cmake_fails(compile_mock, solc_binary, cwd):
    compile_mock.raise_on("cmake")
    with pytest.raises(SolcInstallationError):
        solcx.compile_solc("latest")

    assert os.getcwd() == cwd


@pytest.mark.skipif("sys.platform == 'win32'")
def test_compile_install_make_fails(compile_mock, solc_binary, cwd):
    compile_mock.raise_on("make")
    with pytest.raises(SolcInstallationError):
        solcx.compile_solc("latest")

    assert os.getcwd() == cwd



import functools

import pytest
from packaging.version import Version

import solcx
from solcx.exceptions import SolcNotInstalled, UnsupportedVersionError


@pytest.mark.parametrize(
    "pragma,expected",
    [
        ("0.4.11", "0.4.11"),
        ("=0.4.11", "0.4.11"),
        ("^0.4.11", "0.4.25"),
        (">=0.4.0<0.4.25", "0.4.11"),
        (">=0.4.2", "1.2.3"),
        (">=0.4.2<0.5.5", "0.5.4"),
        ("^0.4.2 || 0.5.5", "0.4.25"),
        ("^0.4.2 || >=0.5.4<0.7.0", "0.5.7"),
        ("0.4.2 || >=0.5.4<0.7.0", "0.5.7"),
        ("^0.5.00", "0.5.7"),
        ("=0.4.11 >=0.4.0 <0.5.0 >=0.4.2 <0.5.0", "0.4.11"),
    ],
)
def test_set_solc_version_pragma(pragmapatch, pragma, expected):
    set_pragma = functools.partial(solcx.set_solc_version_pragma, check_new=True)
    assert set_pragma(f"pragma solidity {pragma};") == Version(expected)


def test_set_solc_version_pragma_not_installed(pragmapatch):
    set_pragma = functools.partial(solcx.set_solc_version_pragma, check_new=True)
    with pytest.raises(SolcNotInstalled):
        set_pragma("pragma solidity ^0.7.1;")


def test_install_solc_version_pragma(pragmapatch):
    install_pragma = functools.partial(solcx.install_solc_pragma, install=False)
    assert install_pragma("pragma solidity ^0.4.11;") == Version("0.4.26")
    assert install_pragma("pragma solidity >=0.4.0<0.4.25;") == Version("0.4.24")
    assert install_pragma("pragma solidity >=0.4.2;") == Version("0.6.0")
    assert install_pragma("pragma solidity >=0.4.2<0.5.5;") == Version("0.5.3")
    assert install_pragma("pragma solidity ^0.4.2 || 0.5.5;") == Version("0.4.26")
    assert install_pragma("pragma solidity ^0.4.2 || >=0.5.4<0.7.0;") == Version("0.6.0")
    with pytest.raises(UnsupportedVersionError):
        install_pragma("pragma solidity ^0.7.1;")


def test_get_solc_version():
    version = solcx.get_solc_version()
    version_with_hash = solcx.get_solc_version(with_commit_hash=True)

    assert version != version_with_hash
    assert version == Version(version_with_hash.base_version)



import subprocess
from typing import Mapping, cast

import pytest
from packaging.version import Version

import solcx
from solcx.exceptions import UnknownOption, UnknownValue


class PopenPatch:
    def __init__(self):
        self.proc = subprocess.Popen
        self.args = []

    def __call__(self, cmd, **kwargs):
        if cmd[1] == "--version":
            return self.proc(cmd, **kwargs)
        assert cmd[0] == str(solcx.install.get_executable())
        for i in self.args:
            assert i in cmd
        return self.proc(cmd, **kwargs)

    def expect(self, *args):
        self.args = [f"--{i.replace('_', '-')}" for i in args]


@pytest.fixture
def popen(monkeypatch):
    patch = PopenPatch()
    monkeypatch.setattr("subprocess.Popen", patch)
    yield patch


@pytest.fixture(autouse=True)
def setup(all_versions):
    pass


def test_help(popen):
    popen.expect("help")
    solcx.wrapper.solc_wrapper(help=True)


@pytest.mark.parametrize(
    "kwarg",
    [
        "help",
        "version",
        "optimize",
        "gas",
        "asm",
        "asm_json",
        "opcodes",
        "bin",
        "bin_runtime",
        "abi",
        "hashes",
        "userdoc",
        "devdoc",
        "standard_json",
    ],
)
def test_boolean_kwargs(popen, foo_source, kwarg):
    popen.expect(kwarg)
    solcx.wrapper.solc_wrapper(stdin=foo_source, **cast(Mapping, {kwarg: True}))


@pytest.mark.parametrize(
    "kwarg,min_solc",
    [("ast", "0.6.0"), ("clone_bin", "0.5.0"), ("formal", "0.5.0"), ("ast_json", "0.8.2")],
)
def test_removed_kwargs(popen, foo_source, kwarg, min_solc):
    popen.expect(kwarg)
    data = cast(Mapping, {kwarg: True})
    version = solcx.get_solc_version()
    if version >= Version(min_solc):
        with pytest.raises(UnknownOption):
            solcx.wrapper.solc_wrapper(stdin=foo_source, **data)
    else:
        solcx.wrapper.solc_wrapper(stdin=foo_source, **data)


def test_unknown_value(foo_source, all_versions):
    expected = UnknownValue if all_versions >= Version("0.4.21") else UnknownOption
    with pytest.raises(expected):
        solcx.wrapper.solc_wrapper(stdin=foo_source, evm_version="potato")


@pytest.mark.parametrize(
    "kwarg,value",
    [
        ("optimize_runs", 200),
        ("libraries", "libraries:0x1234567890123456789012345678901234567890"),
        ("output_dir", "."),
        ("combined_json", "abi"),
        ("allow_paths", "."),
    ],
)
def test_value_kwargs(popen, foo_source, kwarg, value):
    popen.expect(kwarg)
    data = cast(Mapping, {kwarg: value})
    solcx.wrapper.solc_wrapper(stdin=foo_source, **data)



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from packaging.version import Version

from solcx import wrapper
from solcx.exceptions import ContractsNotFound, SolcError
from solcx.install import get_executable


def get_solc_version(with_commit_hash: bool = False) -> Version:
    """
    Get the version of the active ``solc`` binary.

    Args:
      with_commit_hash (bool): If ``True``, the commit hash is included
        within the version. Defaults to ``False``.

    Returns:
      Version: solc version
    """
    solc_binary = get_executable()
    return wrapper.get_solc_version(solc_binary, with_commit_hash=with_commit_hash)


def compile_source(
    source: str,
    output_values: Optional[List] = None,
    import_remappings: Optional[Union[Dict, List, str]] = None,
    base_path: Optional[Union[Path, str]] = None,
    allow_paths: Optional[Union[List, Path, str]] = None,
    output_dir: Optional[Union[Path, str]] = None,
    overwrite: bool = False,
    evm_version: Optional[str] = None,
    revert_strings: Optional[Union[List, str]] = None,
    metadata_hash: Optional[str] = None,
    metadata_literal: bool = False,
    optimize: bool = False,
    optimize_runs: Optional[int] = None,
    optimize_yul: Optional[bool] = None,
    no_optimize_yul: Optional[bool] = None,
    yul_optimizations: Optional[int] = None,
    solc_binary: Optional[Union[str, Path]] = None,
    solc_version: Optional[Union[str, Version]] = None,
    allow_empty: bool = False,
    **kwargs: Any,
) -> Dict:
    """
    Compile a Solidity contract.

    Compilation is handled via the `--combined-json` flag. Depending on the solc
    version used, some keyword arguments may not be available.

    Args:
      source (str): Solidity contract to be compiled.
      output_values (Optional[List]): Compiler outputs to return.
        Valid options depend on the version of ``solc``. If not given,
        all possible outputs for the active version are returned.
      import_remappings (Optional[Union[Dict, List, str]]): Path remappings.
        May be given as a string or list of strings formatted as
        ``"prefix=path"``,or a dict of ``{"prefix": "path"}``.
      base_path (Optional[Union[Path, str]]): Use the given path as the
        root of the source tree instead of the root of the filesystem.
      allow_paths (Optional[Union[List, Path, str]]): A path, or list of
        paths, to allow for imports.
      output_dir (Optional[str]): Creates one file per component and
        contract/file at the specified directory.
      overwrite (bool): Overwrite existing files, used in combination with
        ``output_dir``.
      evm_version (Optional[str]): Select the desired EVM version. Valid
        options depend on the ``solc`` version.
      revert_strings (Optional[List[str]]): Strip revert (and require)
        reason strings or add additional debugging information.
      metadata_hash (Optional[str]): Choose hash method for the bytecode
        metadata or disable it.
      metadata_literal (bool): Store referenced sources as literal data
        in the metadata output.
      optimize (bool): Enable bytecode optimizer. Defauls to ``False``.
      optimize_runs (Optional[int]): Set for how many contract runs to
        optimize. Lower values will optimize more for initial deployment
        cost, higher values will optimize more for high-frequency usage.
      optimize_yul (Optional[bool]): Enable the Yul optimizer. Defaults to
        ``None``.
      no_optimize_yul (Optional[bool]): Disable the Yul optimizer. Defaults to
        ``None``.
      yul_optimizations (Optional[int]): Force yul optimizer to use the
        specified sequence of optimization steps instead of the built-in one.
      solc_binary (Optional[Union[str, Path]]): Path of the `solc` binary to
        use. If not given, the currently active version is used, as set by
        :meth:`solcx.set_solc_version`.
      solc_version (Optional[Union[str, Version]]): ``solc`` version to use. If not given,
        the currently active version is used. Ignored if `solc_binary` is also
        given.
      allow_empty (bool): If ``True``, do not raise when no compiled contracts
        are returned.

    Returns:
      Dict: Compiler output. The source file name is given as ``<stdin>``.
    """
    return _compile_combined_json(
        solc_binary=solc_binary,
        solc_version=solc_version,
        stdin=source,
        output_values=output_values,
        import_remappings=import_remappings,
        base_path=base_path,
        allow_paths=allow_paths,
        output_dir=output_dir,
        overwrite=overwrite,
        evm_version=evm_version,
        revert_strings=revert_strings,
        metadata_hash=metadata_hash,
        metadata_literal=metadata_literal,
        optimize=optimize,
        optimize_runs=optimize_runs,
        optimize_yul=optimize_yul,
        no_optimize_yul=no_optimize_yul,
        yul_optimizations=yul_optimizations,
        allow_empty=allow_empty,
        **kwargs,
    )


def compile_files(
    source_files: Union[List, Path, str],
    output_values: Optional[List] = None,
    import_remappings: Optional[Union[Dict, List, str]] = None,
    base_path: Optional[Union[Path, str]] = None,
    allow_paths: Optional[Union[List, Path, str]] = None,
    output_dir: Optional[Union[Path, str]] = None,
    overwrite: bool = False,
    evm_version: Optional[str] = None,
    revert_strings: Optional[Union[List, str]] = None,
    metadata_hash: Optional[str] = None,
    metadata_literal: bool = False,
    optimize: bool = False,
    optimize_runs: Optional[int] = None,
    optimize_yul: Optional[bool] = None,
    no_optimize_yul: Optional[bool] = None,
    yul_optimizations: Optional[int] = None,
    solc_binary: Optional[Union[str, Path]] = None,
    solc_version: Optional[Union[str, Version]] = None,
    allow_empty: bool = False,
    **kwargs: Any,
) -> Dict:
    """
    Compile one or more Solidity source files.

    Compilation is handled via the ``--combined-json`` flag. Depending on the
    solc version used, some keyword arguments may not be available.

    Args:
      source_files (Union[List, Path, str]): Path, or list of paths, of Solidity
        source files to be compiled.
      output_values (Optional[List]): Compiler outputs to return. Valid options
        depend on the version of ``solc``. If not given, all possible outputs for
        the active version are returned.
      import_remappings (Optional[Union[Dict, List, str]]): Path remappings. May
        be given as a string or list of strings formatted as ``"prefix=path"``,
        or a dict of ``{"prefix": "path"}``.
      base_path (Optional[Union[Path, str]]): Use the given path as the root of
        the source tree instead of the root of the filesystem.
      allow_paths (Optional[Union[List, Path, str]]): A path, or list of paths,
        to allow for imports.
      output_dir (Optional[Union[Path, str]]): Creates one file per component
        and contract/file at the specified directory.
      overwrite (bool): Overwrite existing files, used in combination with
        ``output_dir``.
      evm_version (Optional[str]): Select the desired EVM version. Valid
        options depend on the ``solc`` version.
      revert_strings (Optional[Union[List, str]]): Strip revert (and require)
        reason strings or add additional debugging information.
      metadata_hash (Optional[str]): Choose hash method for the bytecode
        metadata or disable it.
      metadata_literal (Optional[bool]): Store referenced sources as literal
        data in the metadata output.
      optimize (bool): Enable bytecode optimizer. Defaults to ``False``.
      optimize_runs (Optional[int]): Set for how many contract runs to optimize.
        Lower values will optimize more for initial deployment cost, higher
        values will optimize more for high-frequency usage.
      optimize_yul (Optional[bool]): Enable the Yul optimizer. Defaults to
        ``None``.
      no_optimize_yul (Optional[bool]): Disable the Yul optimizer. Defaults to
        ``None``.
      yul_optimizations (Optional[int]): Force yul optimizer to use the
        specified sequence of optimization steps instead of the built-in one.
      solc_binary (Optional[Union[str, Path]]): Path of the `solc` binary to
        use. If not given, the currently active version is used, as set by
        :meth:`solcx.set_solc_version`.
      solc_version (Optional[Union[str, Version]]): ``solc`` version to use. If not given,
        the currently active version is used. Ignored if `solc_binary` is also given.
      allow_empty (bool): If ``True``, do not raise when no compiled contracts are
        returned. defaults to ``False``.

    Returns:
      Dict: Compiler output
    """
    return _compile_combined_json(
        solc_binary=solc_binary,
        solc_version=solc_version,
        source_files=source_files,
        output_values=output_values,
        import_remappings=import_remappings,
        base_path=base_path,
        allow_paths=allow_paths,
        output_dir=output_dir,
        overwrite=overwrite,
        evm_version=evm_version,
        revert_strings=revert_strings,
        metadata_hash=metadata_hash,
        metadata_literal=metadata_literal,
        optimize=optimize,
        optimize_runs=optimize_runs,
        optimize_yul=optimize_yul,
        no_optimize_yul=no_optimize_yul,
        yul_optimizations=yul_optimizations,
        allow_empty=allow_empty,
        **kwargs,
    )


def _get_combined_json_outputs(solc_binary: Optional[Union[Path, str]] = None) -> str:
    if solc_binary is None:
        solc_binary = get_executable()

    help_str = wrapper.solc_wrapper(solc_binary=solc_binary, help=True)[0].split("\n")
    combined_json_args = next(i for i in help_str if i.startswith("  --combined-json"))
    return combined_json_args.split(" ")[-1]


def _parse_compiler_output(stdoutdata: str) -> Dict:
    output = json.loads(stdoutdata)

    contracts = output.get("contracts", {})
    sources = output.get("sources", {})

    for path_str, data in contracts.items():
        if "abi" in data and isinstance(data["abi"], str):
            data["abi"] = json.loads(data["abi"])
        key = path_str.rsplit(":", maxsplit=1)[0]
        if "AST" in sources.get(key, {}):
            data["ast"] = sources[key]["AST"]

    return contracts


def _compile_combined_json(
    output_values: Optional[List] = None,
    solc_binary: Union[str, Path, None] = None,
    solc_version: Optional[Union[Version, str]] = None,
    output_dir: Union[str, Path, None] = None,
    overwrite: Optional[bool] = False,
    allow_empty: Optional[bool] = False,
    **kwargs: Any,
) -> Dict:
    solc_binary = get_executable(version=solc_version) if solc_binary is None else solc_binary

    if output_values is None:
        combined_json = _get_combined_json_outputs(solc_binary)
    else:
        combined_json = ",".join(output_values)

    if output_dir:
        output_dir = Path(output_dir)
        if output_dir.is_file():
            raise FileExistsError("`output_dir` must be as a directory, not a file")
        if output_dir.joinpath("combined.json").exists() and not overwrite:
            target_path = output_dir.joinpath("combined.json")
            raise FileExistsError(
                f"Target output file {target_path} already exists - use overwrite=True to overwrite"
            )

    stdoutdata, stderrdata, command, proc = wrapper.solc_wrapper(
        solc_binary=solc_binary,
        combined_json=combined_json,
        output_dir=output_dir,
        overwrite=overwrite,
        **kwargs,
    )

    if output_dir:
        output_path = Path(output_dir).joinpath("combined.json")
        if stdoutdata:
            output_path.parent.mkdir(parents=True, exist_ok=True)
            with output_path.open("w") as fp:
                fp.write(stdoutdata)
        else:
            with output_path.open() as fp:
                stdoutdata = fp.read()

    contracts = _parse_compiler_output(stdoutdata)

    if not contracts and not allow_empty:
        raise ContractsNotFound(
            command=command,
            return_code=proc.returncode,
            stdout_data=stdoutdata,
            stderr_data=stderrdata,
        )
    return contracts


def compile_standard(
    input_data: Dict,
    base_path: Optional[Union[str, Path]] = None,
    allow_paths: Optional[Union[List, Path, str]] = None,
    output_dir: Optional[str] = None,
    overwrite: bool = False,
    solc_binary: Optional[Union[str, Path]] = None,
    solc_version: Optional[Union[str, Version]] = None,
    allow_empty: bool = False,
) -> Dict:
    """
    Compile Solidity contracts using the JSON-input-output interface.

    See the Solidity documentation for details on the expected JSON input and output
    formats.

    Args:
      input_data (Dict): Compiler JSON input.
      base_path (Optional[Union[str, Path]]): Use the given path as the root of the
        source tree instead of the root of the filesystem.
      allow_paths (Optional[Union[List, Path, str]]): A path, or list of paths, to
        allow for imports.
      output_dir (Optional[str]): Creates one file per component and contract/file
        at the specified directory.
      overwrite (bool): Overwrite existing files, used in combination with
        ``output_dir``.
      solc_binary (Optional[Union[str, Path]]): Path of the `solc` binary to use.
        If not given, the currently active version is used, as set by
        :meth:`solcx.set_solc_version`.
      solc_version (Optional[Union[str, Version]]): ``solc`` version to use. If not given,
        the currently active version is used. Ignored if ``solc_binary`` is also given.
      allow_empty (bool): If ``True``, do not raise when no compiled contracts are returned.
        Defaults to ``False``.

    Returns:
      Dict: Compiler JSON output.
    """
    if not input_data.get("sources") and not allow_empty:
        raise ContractsNotFound(
            "Input JSON does not contain any sources",
            stdin_data=json.dumps(input_data, sort_keys=True, indent=2),
        )

    solc_binary = get_executable(version=solc_version) if solc_binary is None else solc_binary
    stdoutdata, stderrdata, command, proc = wrapper.solc_wrapper(
        solc_binary=solc_binary,
        stdin=json.dumps(input_data),
        standard_json=True,
        base_path=base_path,
        allow_paths=allow_paths,
        output_dir=output_dir,
        overwrite=overwrite,
    )

    compiler_output = json.loads(stdoutdata)
    if "errors" in compiler_output:
        has_errors = any(error["severity"] == "error" for error in compiler_output["errors"])
        if has_errors:
            error_message = "\n".join(
                tuple(
                    error["formattedMessage"]
                    for error in compiler_output["errors"]
                    if error["severity"] == "error"
                )
            )
            raise SolcError(
                error_message,
                command=command,
                return_code=proc.returncode,
                stdin_data=json.dumps(input_data),
                stdout_data=stdoutdata,
                stderr_data=stderrdata,
                error_dict=compiler_output["errors"],
            )
    return compiler_output


def link_code(
    unlinked_bytecode: str,
    libraries: Dict,
    solc_binary: Optional[Union[str, Path]] = None,
    solc_version: Optional[Version] = None,
) -> str:
    """
    Add library addresses into unlinked bytecode.

    Args:
      unlinked_bytecode (str): Compiled bytecode containing one or
        more library placeholders.
      libraries (Dict): Library addresses given as
        ``{"library name": "address"}``.
      solc_binary (Optional[Union[str, Path]]): Path of the ``solc``
        binary to use. If not given, the currently active
        version is used, as set by :meth:`solcx.set_solc_version`.
      solc_version (Optional[Version]): `solc` version to use. If
        not given, the currently active version is used. Ignored if
        ``solc_binary`` is also given.

    Returns:
      str: Linked bytecode
    """
    solc_binary = get_executable(version=solc_version) if solc_binary is None else solc_binary
    library_list = [f"{name}:{address}" for name, address in libraries.items()]

    stdoutdata = wrapper.solc_wrapper(
        solc_binary=solc_binary, stdin=unlinked_bytecode, link=True, libraries=library_list
    )[0]

    return stdoutdata.replace("Linking completed.", "").strip()



import pytest


class WrapperMock:
    """
    Simple mock for solcx.wrapper.solc_wrapper
    """

    def __call__(self, **kwargs):
        for key, value in self.kwargs.items():
            assert kwargs[key] == value
        return '{"contracts":{}}', "", [], 0

    def expect(self, **kwargs):
        self.kwargs = kwargs


@pytest.fixture
def wrapper_mock(monkeypatch):
    mock = WrapperMock()
    monkeypatch.setattr("solcx.wrapper.solc_wrapper", mock)
    yield mock


class CompileCombinedJsonMock:
    """
    Simple mock for solcx.main._compile_combined_json
    """

    def __call__(self, **kwargs):
        for key, value in self.kwargs.items():
            assert kwargs[key] == value
        return {}

    def expect(self, **kwargs):
        self.kwargs = kwargs


@pytest.fixture
def compile_combined_json_mock(monkeypatch):
    mock = CompileCombinedJsonMock()
    monkeypatch.setattr("solcx.main._compile_combined_json", mock)
    yield mock



from solcx import wrapper
from solcx.install import (
    compile_solc,
    get_compilable_solc_versions,
    get_installable_solc_versions,
    get_installed_solc_versions,
    get_solcx_install_folder,
    import_installed_solc,
    install_solc,
    install_solc_pragma,
    set_solc_version,
    set_solc_version_pragma,
)
from solcx.main import compile_files, compile_source, compile_standard, get_solc_version, link_code

__all__ = [
    "compile_files",
    "compile_solc",
    "compile_source",
    "compile_standard",
    "get_compilable_solc_versions",
    "get_installable_solc_versions",
    "get_installed_solc_versions",
    "get_solc_version",
    "get_solcx_install_folder",
    "import_installed_solc",
    "install_solc",
    "install_solc_pragma",
    "link_code",
    "set_solc_version",
    "set_solc_version_pragma",
    "wrapper",
]
